import sys, os, string, re
import socket, threading, getopt
from SocketServer import BaseRequestHandler, TCPServer

OUTBOUND = 1
INBOUND  = 0

class LoggingServer( TCPServer ):

    _verbosity          = 2
    _listen_host        = 'localhost'
    _listen_port        = 9999
    _logfile_directory  = '/tmp'
    _logfile_prefix     = 'zft'
    _record_inbound     = 0

    def __init__( self, args, filter_class ):
        self.parseOptions( args )
        address = ( self._listen_host, self._listen_port )
        TCPServer.__init__( self, address, filter_class )

    def printUsage( self, msg=None ):
        """
            Print command-line help.
        """
        print """\

LoggingProxy [-?vqiI] [-h listen_host] [-p listen_port] \\
                     [-l log_dir] [-f log_prefix]

    -?, --help                      Print this help message

    -v, --verbose                   Increment verbosity (default is '1')

    -q, --quiet                     Set verbosity to '0'

    -i, --record-inbound            Record inbound data

    -I, --no-record-inbound         Don't record inbound data (default)

    -h, --listen-host               Host address to listen on
                                    (default '%(LISTEN_HOST)s')

    -p, --listen-port               Port number to listen on
                                    (default '%(LISTEN_PORT)s')

    -l, --logfile-directory         Directory to write log files
                                    (default '%(LOGFILE_DIRECTORY)s')

    -f, --logfile-prefix            Prefix for log file names
                                    (default '%(LOGFILE_PREFIX)s')

%(MESSAGE)s
        """ % ( { 'LISTEN_HOST'         : self._listen_host
                , 'LISTEN_PORT'         : self._listen_port
                , 'LOGFILE_DIRECTORY'   : self._logfile_directory
                , 'LOGFILE_PREFIX'      : self._logfile_prefix
                , 'MESSAGE'             : msg and 'Error: %s' % msg or ''
                } )
        sys.exit( 1 )

    def parseOptions( self, args ):
        """
            Parse command line options.
        """
        verbosity = self._verbosity
        record_inbound = self._record_inbound
        listen_host = listen_port = logfile_directory = logfile_prefix = None

        try:
            opts, ignored = getopt.getopt( args
                                         , "?vqiIh:p:l:f:"
                                         , [ 'help'
                                           , 'verbose'
                                           , 'quiet'
                                           , 'record-inbound'
                                           , 'no-record-inbound'
                                           , 'listen-host='
                                           , 'listen-port='
                                           , 'logfile-directory='
                                           , 'logfile-prefix='
                                           ]
                                         )
        except getopt.GetoptError, msg:
            self.printUsage( msg=msg)

        for o, v in opts:

            if o == '-?' or o == '--help':
                self.printUsage()

            if o == '-v' or o == '--verbose':
                verbosity = verbosity + 1

            if o == '-q' or o == '--quiet':
                verbosity = 0

            if o == '-i' or o == '--record-inbound':
                record_inbound = 1

            if o == '-I' or o == '--no-record-inbound':
                record_inbound = 0

            if o == '-h' or o == '--listen-host':
                listen_host = v

            if o == '-p' or o == '--listen-port':
                listen_port = v

            if o == '-l' or o == '--logfile-directory':
                logfile_directory = v

            if o == '-f' or o == '--logfile-prefix':
                logfile_prefix = v

        self._verbosity = verbosity
        self._record_inbound = record_inbound

        if listen_host is not None:
            self._listen_host = listen_host

        if listen_port is not None:
            self._listen_port = int( listen_port )

        if logfile_directory is not None:
            self._logfile_directory = logfile_directory

        if logfile_prefix is not None:
            self._logfile_prefix = logfile_prefix

    def process_request( self, request, client_address ):
        """
            Start a new thread to process the request.
        """
        t = threading.Thread( target=self.finish_request
                            , args=( request, client_address ) )
        t.start()



handler_count = 0
hc_lock = threading.RLock()

def incrementHC():
    global handler_count
    hc_lock.acquire()
    try:
        handler_count = handler_count + 1
        return handler_count
    finally:
        hc_lock.release()

class LoggingFilter( BaseRequestHandler ):
    """
    """

    def handle( self ):
        """
            Organise the data flows.
        """
        LSW = self._getWrapperClass()
        count = incrementHC()
        try:
            directory = self._getDirectory()
            prefix = self._getPrefix()
            client = LSW( skt=self.request
                        , directory=directory
                        , prefix=prefix
                        , count=count
                        , direction=INBOUND
                        , suppress_logging=not self.server._record_inbound
                        , verbosity=self.server._verbosity
                        )
            server = LSW( skt=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        , directory=directory
                        , prefix=prefix
                        , count=count
                        , direction=OUTBOUND
                        , verbosity=self.server._verbosity
                        )
            self.doRequest( client, server )
            self.doTransfer( client, server )
        except:
            self._logTraceback()
        self.close( client, server )

    def doRequest( self
                 , client
                 , server
                 , HTTP_REQUEST = re.compile( r'^\s*(\S+)\s+'
                                            + r'http://([^\s/:]+)(:(\d+))'
                                            + r'?(\S*)\s*(\S+)\s*([\n\r]+)$'
                                            )
                 ):
        """
            Process the request line (first line from the client).
        """
        req = client.readln()
        match = HTTP_REQUEST.search( req )

        if not match:
            raise IOError, 'Unexpected Url format: %s' % req

        cmd, name, junk, port, path, protocol, cr = match.groups()

        if not port:
            port = "80"

        address = ( name, int( port ) )
        req = string.join( ( cmd, ' ', path, ' ', 'HTTP/1.0', '\r\n' ), '' )

        server.connect( address )
        server.send( req ) # send the modified request
        server.log( 'Request: %s' % req )
        return name

    def doTransfer( self, client, server ):
        """
            Transfer data back and forth.
        """
        length = client.forwardHeaders( server, OUTBOUND )

        if length:
            client.forwardBinary( server, OUTBOUND, length )

        length = server.forwardHeaders( client, INBOUND )

        server.forwardBinary( client, INBOUND, length )

    def close( self, client, server ):
        """
            Close the sockets.
        """
        try:
            client.close()
        except:
            pass
        try:
            server.close()
        except:
            pass

    #
    #   Derived filters should override these to adapt to a different UI.
    #
    def _getWrapperClass( self ):
        """
            Return the class from which to make socket wrappers.
        """
        return LoggingServerWrapper

    def _getDirectory( self ):
        """
            Return the directory path used for storing logfiles.
        """
        return self.server._logfile_directory

    def _getPrefix( self ):
        """
            Return the prefixed used for naming logfiles.
        """
        return self.server._logfile_prefix

    def _logTraceback( self ):
        """
            Log an error.
        """
        import traceback
        traceback.print_exc()


class LoggingServerWrapper:
    """
        Extends SW to log outbound headers and data, plus returned headers.
    """

    def __init__( self
                , skt
                , directory
                , prefix
                , count
                , direction
                , suppress_logging=0
                , verbosity=0
                ):

        self.skt = skt
        self.inBuffer = ''
        self.chunkSize = 8192
        self.readAll = 0 # no more data to read
        self.verbosity = verbosity
        if suppress_logging:
            self.teeFile = None
        else:
            teeFileName = '%s/%s_%03d.%s' % ( directory
                                            , prefix
                                            , count
                                            , direction and 'out' or 'in'
                                            )
            self.teeFile = open( teeFileName, 'w' )

    def readln( self
              , LINE=re.compile( r'^([^\n]*\n\r?)(.*)', re.S )
              ):
        """
            Get a line from the buffer (reading if necessary).
            Return None on error.
        """
        while 1:
            match = LINE.search( self.inBuffer )
            if match:
                break
            else: # no match, but may have all the data anyway
                if self.readAll: # so return what there is
                    if self.inBuffer == '':
                        return None # error?
                    line = self.inBuffer
                    self.inBuffer = ''
                    return line
                else: # or read more
                    self.extendBuffer()
        line, self.inBuffer = match.groups() # save remainder in buffer
        return line

    def extendBuffer( self ):
        """
            Read more from the socket, extending the buffer.
        """
        if not self.readAll:
            more = self.skt.recv( self.chunkSize )
            if not more:
                self.readAll = 1
            else:
                self.inBuffer = string.join((self.inBuffer, more), '')

    def connect( self, address ):
        """
            Connect the underlying socket.
        """
        self.skt.connect( address )

    def send( self, string ):
        """
            Send to the underlying socket.
        """
        self.skt.send( string )
        if self.teeFile:
            self.teeFile.write( string )

    def log( self, msg ):
        if self.verbosity:
            sys.stderr.write( '* %s\n' % msg )

    def forwardHeaders( self
                      , skt
                      , direction
                      , CONTENT_LENGTH = re.compile(r'Content-length:\s*(\d+)')
                      , EMPTY = re.compile( r'^\s*\n\r?$' )
                      ):
        """
            Copy headers from this socket to the given destination.
        """
        length = 0
        sent_connection_close = 0
        while 1:

            line = self.readln()

            if line == None:
                break # connection closed

            if line and line[0] not in ( ' ', '\t', '\r', '\n' ):

                try:
                    header_name, rest = string.split( line, ':', 1 )

                except ValueError:  # split -> unpack failed.
                    pass

                else:

                    if header_name.lower() in ( 'connection'
                                              , 'proxy-connection'
                                              , 'keep-alive'
                                              ):
                        if not sent_connection_close:
                            skt.send( 'Connection: close' )
                            sent_connection_close = 1
                        continue

            match = CONTENT_LENGTH.search( line )

            if match:
                length = int( match.group( 1 ) )

            skt.send( line )
            self.log( '%s header: %s' % ( direction and 'out' or 'in'
                                        , line ) )

            if EMPTY.search( line ):
                break

        return length

    def forwardBinary( self, skt, direction, length=0 ):
        """
            Extend base to record actual data sent.
        """
        count = length > 0
        while 1:
            if self.inBuffer:
                skt.send( self.inBuffer )
                size = len( self.inBuffer )
                self.log( '%s data: %d' % ( direction and 'out' or 'in'
                                          , size ) )
                self.inBuffer = ''
                if count:
                    length = length - size
                    if length <= 0:
                        break
            self.extendBuffer()
            if self.readAll:
                break

    def close(self):
        """
            Close the underlying socket.
        """
        self.skt.close()

if __name__ == '__main__':

    import sys

    server = LoggingServer( sys.argv[1:], LoggingFilter )
    server.serve_forever()
