"""
    LoggingProxy request post-processor.
"""

import sys, string, re
import getopt, glob, rfc822, urllib

SCENARIO_SKELETON = """\
[Scenario]
Title : Enter scenario title here
Use_case: Enter use case name here
"""

REQUEST_SKELETON = """
[Request_%(REQUEST_NUM)03d]
HTTP_Verb: %(HTTP_VERB)s
HTTP_Version: %(HTTP_VERSION)s
URL: %%(portal_url)s/%%(site_path)s%(REQUEST_URI)s
"""

AUTHENTICATION_SKELETON = """\
Authentication: %%(userid)s:%%(password)s
"""

HEADER_SKELETON = """\
Header_%(HEADER_NUM)d: %(HEADER_NAME)=%(HEADER_VALUE)s
"""

COOKIE_SKELETON = """\
Cookie_%(COOKIE_NUM)d: %(COOKIE_VALUE)s
"""

FIELD_SKELETON = """\
Field_%(FIELD_NUM)d: %(FIELD_NAME)s=%(FIELD_VALUE)s
"""

RESULT_SKELETON = """\
Expected_Result: %(RESULT_CODE)d
"""

DEFAULT_SKELETON = """
[DEFAULT]
userid: userid
password: password
"""

class ScenarioGenerator:
    """
        Convert a series of HTTP requests files (as generated by
        LoggingProxy) into a scenario file (to be consumed by
        FT_Runner).
    """
    _verbosity          = 1
    _capture_cookies    = 0
    _logfile_directory  = '/tmp'
    _logfile_prefix     = 'zft'
    _output_file        = None
    _exclude_patterns   = []
    _exclude_file       = None
    _exclude_regex      = None

    def __init__( self, args ):

        self.parseOptions( args )

    def printUsage( self, msg=None ):
        """
            Dump a help message, and bail out.
        """
        sys.stderr.write( """
  python ScenarioGenerator.py [-?vq] \\
                     [-l log_dir] [-f log_prefix] [-o file] \\
                     [-x pattern] [-X file]

    -?, --help                      Print this help message

    -v, --verbose                   Increment verbosity (default is '1')

    -q, --quiet                     Set verbosity to '0'

    -c, --caputure-cookies          Capture cookies

    -C, --no-capture-cookies        Don't capture cookies (default)

    -l, --logfile-directory         Directory from which to read log files
                                    (default '%(LOGFILE_DIRECTORY)s')

    -f, --logfile-prefix            Prefix for log file names
                                    (default '%(LOGFILE_PREFIX)s')

    -o, --output-file               Write to 'file', instead of default
                                    (%(LOGFILE_PREFIX)s.zft).  Use '-'
                                    to write to stdout.

    -x, --exclude-pattern           Exclude requests which match 'pattern'
                                    (e.g., to suppress stylesheet or images).

    -X, --exclude-file              Exclude requests which match any pattern
                                    read from 'file' (one pattern per line).

%(MESSAGE)s\n""" % { 'LOGFILE_DIRECTORY'    : self._logfile_directory
                   , 'LOGFILE_PREFIX'       : self._logfile_prefix
                   , 'MESSAGE'              : msg or ''
                   } )
        sys.exit( 1 )

    def parseOptions( self, args ):
        """
            Parse command-line options.
        """
        verbosity = self._verbosity
        capture_cookies = self._capture_cookies
        logfile_directory = logfile_prefix = output_file = None
        exclude_file = portal_url = site_path = None

        try:
            opts, ignored = getopt.getopt( args
                                         , "?vqcCl:f:o:x:X:"
                                         , [ 'help'
                                           , 'verbose'
                                           , 'quiet'
                                           , 'capture-cookies'
                                           , 'no-capture-cookies'
                                           , 'logfile-directory='
                                           , 'logfile-prefix='
                                           , 'output-file'
                                           , 'exclude-pattern'
                                           , 'exclude-file'
                                           ]
                                         )
        except getopt.GetoptError, msg:
            self.printUsage( msg=msg)

        for o, v in opts:

            if o == '-?' or o == '--help':
                self.printUsage()

            if o == '-v' or o == '--verbose':
                verbosity = verbosity + 1

            if o == '-q' or o == '--quiet':
                verbosity = 0

            if o == '-c' or o == '--capture-cookies':
                capture_cookies = 1

            if o == '-C' or o == '--no-capture-cookies':
                capture_cookies = 0

            if o == '-l' or o == '--logfile-directory':
                logfile_directory = v

            if o == '-f' or o == '--logfile-prefix':
                logfile_prefix = v

            if o == '-o' or o == '--output-file':
                output_file = v

            if o == '-x' or o == '--exclude-pattern':
                self._addExcludePattern( v )

            if o == '-X' or o == '--exclude-file':
                exclude_file = v

        self._verbosity = verbosity
        self._capture_cookies = capture_cookies

        if logfile_directory is not None:
            self._logfile_directory = logfile_directory

        if logfile_prefix is not None:
            self._logfile_prefix = logfile_prefix

        if output_file is not None:
            if output_file == '-':
                self._output_file = sys.stdout
            else:
                self._output_file = open( output_file, 'w' )
        else:
            self._output_file = open( '%s/%s.zft'
                                    % ( self._logfile_directory
                                      , self._logfile_prefix
                                      ), 'w' )

        if exclude_file is not None:
            self._exclude_file = exclude_file

    def _log( self, msg, level ):
        """
            Write a note to stderr (if verbosity enabled).
        """
        if level <= self._verbosity:
            sys.stderr.write( "%s\n" % msg )

    def _print( self, fmt, **kw ):
        """
            Dump the appropriately-formatted values to our output
            file.
        """
        self._output_file.write( fmt % kw )

    def _handleCookie( self, value, parms ):
        """
            Dump a cookie value.
        """
        if self._capture_cookies:
            count = parms.get( 'cookie_count', 0 ) + 1
            parms[ 'cookie_count' ] = count
            self._print( COOKIE_SKELETON
                       , COOKIE_NUM=count
                       , COOKIE_VALUE=value
                       )

    def _handleAuthentication( self, value, parms ):
        """
            Dump a placeholder for authentication, and record the
            fact that we saw it (so 'processScenario' can write the
            '[DEFAULT]' section).
        """
        parms[ 'saw_authentication' ] = 1
        self._print( AUTHENTICATION_SKELETON )

    def _handleContentType( self, value, parms ):
        """
            Record the content type for later processing.
        """
        parms[ 'content_type' ] = value

    HEADER_HANDLERS =\
    { 'cookie'          : _handleCookie
    , 'authentication'  : _handleAuthentication
    , 'content-type'    : _handleContentType
    }

    def _handleURLEncoded( self, data ):
        """
            Emit Field_## values for data, which is encoded using
            'application/x-www-form-urlencoded'.
        """
        count = 0
        for key_val in string.split( data, '&' ):
            k, v = map( urllib.unquote_plus, string.split( key_val, '=' ) )
            count = count + 1
            self._print( FIELD_SKELETON
                       , FIELD_NUM=count
                       , FIELD_NAME=k
                       , FIELD_VALUE=v
                       )

    def _handleFormdata( self, data ):
        """
            Emit Field_## values for data, which is encoded using
            'multipart/formdata'.
        """
        raise NotImplemented    # TODO:  Implement me!

    CONTENT_HANDLERS =\
    { 'application/x-www-form-urlencoded'   : _handleURLEncoded
    , 'multipart/formdata'                  : _handleFormdata
    }

    def _addExcludePattern( self, pattern ):
        """
            Add a pattern to our list of excluded patterns.
        """
        self._exclude_patterns.append( r'(%s)' % pattern )
        self._exclude_regex = None

    def _getExcludeRegex( self
                        , join=string.join
                        , compile=re.compile
                        ):
        """
            Return a regex which, if matched, indicates that we should
            skip the file.
        """
        if self._exclude_regex is None:

            if self._exclude_file:
                f = open( self._exclude_file )
                for line in f.readlines():
                    line = line.strip()
                    self._addExcludePattern( line )

            if self._exclude_patterns:
                self._exclude_regex = compile( join( self._exclude_patterns
                                                   , '|' ) )
        return self._exclude_regex

    def processFile( self
                   , filename
                   , request_num
                   , parms={}
                   , REQUEST_LINE=re.compile( r'^([^\s]+)\s+'
                                              r'([^\s]+)\s+'
                                              r'([^\s]+)$' )
                   ):
        """
            Process a single request file;  record global context
            in parms.
        """
        self._log( 'Scanning file: %s' % filename, 1 )

        parms[ 'content_type' ] = None
        parms[ 'header_count' ] = 0
        parms[ 'cookie_count' ] = 0

        f = open( filename )
        all_text = f.read()
        f.seek( 0 )

        exclude = self._getExcludeRegex()
        if exclude is not None and exclude.search( all_text ):
            self._log( '** matches exclude regex, skipping', 1 )
            return request_num

        request = string.rstrip( f.readline() )
        match = REQUEST_LINE.match( request )

        if not match:
            self._log( 'Invalid request line: %s' % request, 0 )
            return

        http_verb, uri, http_version = match.groups()

        request_num = request_num + 1

        self._print( REQUEST_SKELETON
                   , REQUEST_NUM=request_num
                   , HTTP_VERB=http_verb
                   , HTTP_VERSION=http_version
                   , REQUEST_URI=uri
                   )

        headers = rfc822.Message( f )
        payload = f.read()
        f.close()

        for k, v in headers.items():
            self._log( 'Header %s: %s' % ( k, v ), 3 )
            handler = self.HEADER_HANDLERS.get( k, None )
            if handler:
                handler( self, v, parms )

        content_type = parms[ 'content_type' ]

        if content_type is not None:
            handler = self.CONTENT_HANDLERS.get( content_type, None )

            if handler is None:
                self._log( "Can't handle content type: %s"
                            % content_type, 0 )
            else:
                handler( self, payload )

        self._print( RESULT_SKELETON
                   , RESULT_CODE=200    # TODO:  Read from response?
                   )

        return request_num

    def processScenario( self ):
        """
            Read all files in '_logfile_directory' whose prefix
            matches '_logfile_prefix';  create a scenario file
            with one section per request file, dumping to specified
            output file.
        """
        self._print( SCENARIO_SKELETON )

        glob_pattern = '%s/%s_*.out' % ( self._logfile_directory
                                       , self._logfile_prefix )

        request_num = 0
        parms = { 'saw_authentication' : 0 }

        for filename in glob.glob( glob_pattern ):
            request_num = self.processFile( filename, request_num, parms )

        if parms[ 'saw_authentication' ]:
            self._print( DEFAULT_SKELETON )

if __name__ == '__main__':

    ScenarioGenerator( sys.argv[1:] ).processScenario()
